/*
 * Kurt Kinetic (New) Road Machine Power Meter
 *
 * (c) Axel (XL) Schwenke, axel.schwenke@gmx.net
 *
 * $Id: main.c 229 2014-01-01 23:29:46Z schwenke $
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <stdio.h>

#include "lcd.h"


/* global definitions */

#define DIA       0.053975      /* roller diameter in m */
#define A1        3.25898       /* linear power koefficient */
#define A3        4.59865E-3    /* cubic power koefficient */
#define AVERAGE   6             /* average over this many ticks */
#define ETA       0.20          /* body efficieny : burned food energy
                                 * vs mechanical output for cycling */

/* global variables */

volatile uint32_t stamp;        /* timestamp of last pulse */
volatile uint16_t pulses;       /* number of pulses in interval */
volatile uint16_t tick;         /* clock tick, unit = 2^16 s */

uint32_t last_stamp;

/*
 * LCD buffer, 4 lines x 20 chars + NUL
 *
 *             -->12345678901234567890<-- */
char line1[21] = "  Power Meter for   ";
char line2[21] = " Kurt Kinetic Fluid ";
char line3[21] = "  Resistance Units  ";
char line4[21] = "V1.0 (c)XL 2013,2014";

/* workout variables */
struct {
    double time;  /* ride time in seconds */
    double dist;  /* ride distance in meters */
    double work;  /* integral work in J = Ws */
    double speed; /* current speed in km/h */
    double power; /* current power in W */
} workout;


/* INTERRUPT SERVICE ROUTINES */

/* Timer 1 Overflow - time keeper */
ISR(TIMER1_OVF_vect)
{
    ++tick;
}

/* Resistance unit rotation impulse */
ISR(TIMER1_CAPT_vect)
{
    stamp = (uint32_t)tick << 16 | ICR1;
    ++pulses;
}


/* SUB ROUTINES */

/* initialize hardware resources */
static void hardware_init(void)
{
    /* pullups enabled on all inputs */
    /* data bus pins are inputs */

    PORTB= 0xFF;
    DDRB=  0;

    PORTC= ~(_BV(LCD_E_PIN));
    DDRC=  (_BV(LCD_RS_PIN) | _BV(LCD_RW_PIN) | _BV(LCD_E_PIN));

    PORTD= 0xFF;
    DDRD=  0;

    /* disable Analog Comparator */

    ACSR = (_BV(ACD));

    /* Timer 1: system clock/8 */

    TCCR1A= 0;
    TCCR1B= (_BV(CS11));

    /* Timer 1: capture irq and
     * Timer 1: overflow irq enabled */

    TIMSK1=  (_BV(ICIE1) | _BV(TOIE1));

    /* enable idle sleep */
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_enable();
}


/* fill display buffer with spaces */
static void clear_display(void)
{
    for (uint8_t i=0; i<LCD_DISP_LEN; i++) {
        line1[i]= ' ';
        line2[i]= ' ';
        line3[i]= ' ';
        line4[i]= ' ';
    }
}


/* reset global variables */
static void reset_workout(void)
{
    workout.time= 0;
    workout.dist= 0;
    workout.work= 0;
    workout.speed= 0;
    workout.power= 0;
}


/* flush the time part of the display buffer to the display
 * replace NULL chars with spaces */
static void flush_time(void)
{
    uint8_t i;
    lcd_gotoxy(0,0);
    for (i=0; i<10; i++)
        lcd_putc(line1[i] ? line1[i] : ' ');
}


/* flush the full display buffer to the display
 * replace NULL chars with spaces */
static void flush_display(void)
{
    uint8_t i;

    lcd_gotoxy(0,0);
    for (i=0; i<LCD_DISP_LEN; i++)
        lcd_putc(line1[i] ? line1[i] : ' ');

    lcd_gotoxy(0,1);
    for (i=0; i<LCD_DISP_LEN; i++)
        lcd_putc(line2[i] ? line2[i] : ' ');

    lcd_gotoxy(0,2);
    for (i=0; i<LCD_DISP_LEN; i++)
        lcd_putc(line3[i] ? line3[i] : ' ');

    lcd_gotoxy(0,3);
    for (i=0; i<LCD_DISP_LEN; i++)
        lcd_putc(line4[i] ? line4[i] : ' ');
}


/* fill ride time into display buffer
 * this overflows after 18:12:15 (hh:mm:ss) */
static void setup_time(uint16_t t)
{
    uint8_t sec= t%60; t/= 60;
    uint8_t min= t%60; t/= 60;
    if (t) {
        snprintf(line1, 11, "%2d:%02d:%02d", t, min, sec);
    } else {
        snprintf(line1, 11, "   %02d:%02d", min, sec);
    }
}


/* independently update the display of the ride time
 * to avoid "jumping" time due to averaging */
void update_time(void)
{
    static uint16_t last_time;

    /* only if moving */
    if (workout.speed) {
        cli();
        uint32_t now= stamp;
        sei();

        uint16_t t= (workout.time + (now-last_stamp)/1000000.0);
        if (t != last_time) {
            setup_time(t);
            flush_time();
            last_time= t;
        }
    }
}


/* fill variables into display buffer
 * and flush the buffer to the display */
static void setup_display(void)
{
    setup_time(workout.time);

    if (workout.dist >= 10000) {
        snprintf(line1+11, 10, "%5.1f km ", workout.dist/1000);
    } else {
        snprintf(line1+11, 10, "%5.2f km ", workout.dist/1000);
    }

    snprintf(line2, 12, "%5.2f km/h ", workout.speed);
    snprintf(line2+11, 10, "%5.2f avg ", workout.dist*3.6/workout.time);

    snprintf(line3, 12, "%5.1f W ", workout.power);
    snprintf(line3+11, 10, "%5.1f avg ", workout.work/workout.time);

    snprintf(line4, 12, "%5.1f kJ", workout.work/1000);

    double cal= workout.work/4186.84/ETA;
    if (cal >= 100) {
        snprintf(line4+11, 10, "%4.0f kcal", cal);
    } else {
        snprintf(line4+11, 10, "%4.1f kcal", cal);
    }
}


void update_workout(void)
{
    cli();
    uint32_t now= stamp;
    uint8_t  p= pulses;
    pulses= 0;
    sei();

    if (p) {
        /* moving */
        if (last_stamp) { /* if we have a start timestamp */
            double delta_t= (now-last_stamp)/1000000.0;     /* s to seconds */
            double delta_s= p * DIA * 3.1415927;
            workout.time+= delta_t;
            workout.dist+= delta_s;
            workout.speed= (delta_s * 3.6) / delta_t;  /* m/s to km/h */
            /* the fluid resistance unit has a simple cubic power curve */
            workout.power= workout.speed *
                (A1 + workout.speed * workout.speed * A3);
            workout.work+= workout.power * delta_t;
        }
        last_stamp= now;
    } else {
        /* stopped */
        workout.speed= 0;
        workout.power= 0;
        last_stamp= 0;
    }
}


int main(void)
{
    hardware_init();
    lcd_init(LCD_DISP_ON);

    /* display version information */
    flush_display();
    _delay_ms(1000);

    clear_display();
    reset_workout();
    sei();

    /* main loop */
    uint16_t last_tick= 0;
    uint8_t  avg_counter= 0;

    while (1) {
        cli();
        uint16_t now= tick;
        sei();

        if (last_tick != now) {
            if (++avg_counter == AVERAGE) {
                avg_counter= 0;
                update_workout();
                setup_display();
                flush_display();
            } else {
                update_time();
            }
            last_tick= now;
        }
        sleep_cpu();
    }
}

